def as_number(values):
    """Receives a series of card values and returns a possible calculation as a numeric value.
    This is tricky because it must deal with non numeric cards (returning -1) and 10's in very specific
    ways."""
    all_10s = True
    for i in values:
        if i != 10:
            all_10s = False
            break
    if all_10s:
        return 10 ** len(values)

    total = 0

    for value in values:

        if value > 10:
            return -1

        total *= 10

        if value != 10:
            total += int(value)
        else:
            # value is 10, first digit is one
            if total == 0:
                total += 1
                # otherwise, ignore the 0

    return total



from generativemagic.checks.poker import IsAcesHand
from generativemagic.effect import Effect


class AcesHand(Effect):

    def apply(self, sequence):
        result = IsAcesHand([4, 5], True).get_first_ace(sequence)
        raise Exception(f"To be implemented at {result}")

    def __repr__(self):
        return f"AcesHand()"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return IsAcesHand([4, 5], True).check(deck)



import logging
from asyncio import Queue
from queue import SimpleQueue
from typing import List, Callable, NewType, Generator, Tuple, Dict

import numpy as np

from generativemagic.decks import simple_deck
from generativemagic.movement import Reorder
from generativemagic.spelling import English
import pandas as pd

ValidationFunction = NewType("ValidationFunction", Callable[[List[int]], bool])
Deck = NewType("Deck", np.array)
LimiterCallback = NewType("LimiterCallback", Callable[[int], bool])


def depth_limiter(k):
    """A limiter on depth k"""
    return lambda depth: depth >= k


def no_limit(depth):
    """No limit on depth"""
    return False


SpaceCombination = NewType("SpaceCombination", Tuple[Deck, List, int])


class Searcher:

    def __init__(self, all_movements: List, extra_limiter: Callable = None):
        self.all_movements = all_movements
        self.extra_limiter = extra_limiter

    def _non_limiting(self, depth):
        return False

    def _bfs_search(self, starting_deck: Deck,
                    validating_function: ValidationFunction,
                    limiter: LimiterCallback):
        valid = validating_function(starting_deck)
        if valid:
            print(f"FOUND IT!! 0")
            return starting_deck, [], valid

        decks_to_evaluate = SimpleQueue()
        decks_to_evaluate.put((starting_deck, [], 0))

        parsed = 0

        while not decks_to_evaluate.empty():

            deck, used, depth = decks_to_evaluate.get()
            if limiter(depth):
                return None

            for movement in self.all_movements:
                if movement in used:
                    continue

                # TODO do not use those global limits
                if not movement.still_can_use(used):
                    continue

                if self.extra_limiter and not self.extra_limiter(movement, used):
                    continue

                resulting_used = used.copy()
                resulting_used.append(movement)

                resulting_deck = movement.apply(deck)
                valid = validating_function(resulting_deck)

                if valid:
                    print(f"FOUND IT!! {depth + 1}")
                    return resulting_deck, resulting_used, valid

                decks_to_evaluate.put(([resulting_deck, resulting_used, depth + 1]))
                parsed += 1

                if parsed % 5000 == 0:
                    msg = ",".join([f"{resulting_deck[10:16]}"]) + ":" + str(depth + 1) + ":" + str(
                        valid)
                    logging.info(f"{parsed} parsed: " + msg)

        return None

    def search(self, starting_deck: Deck,
               validating_function: ValidationFunction,
               limiter: LimiterCallback = no_limit):
        logging.info(f"Starting search with total of #{len(self.all_movements)} movements")
        return self._bfs_search(starting_deck,
                                validating_function,
                                limiter)


def limit_search(validation: ValidationFunction,
                 deck: Deck = None,
                 limits=(1, 2, 3, 4),
                 searcher=None,
                 movements=None):
    if searcher is None:
        searcher = Searcher(movements.values())
    if deck is None:
        deck = simple_deck()

    for limit in limits:
        print(f"Checking limit={limit}")
        result = searcher.search(deck, validation, depth_limiter(limit))
        if result is not None:
            current_deck, used_so_far, valid = result
            print(f"limit {limit} SUCCESS:")
            print(result)
            print(current_deck)
            print(used_so_far)
            print(valid)
            print()
            return True

    return False


class TypeLimiter:
    def __init__(self):
        self.limiters = {}

    def limit(self, who, quantity):
        self.limiters[who] = quantity
        return self

    def check(self, movement, used_so_far):
        if type(movement) != Reorder:
            return True
        # print(self.limiters, movement.simple_name, movement.simple_name not in self.limiters)
        if movement.simple_name not in self.limiters:
            return True
        count = 0
        # print(f"{movement} might be limited")
        for used in used_so_far:
            if movement.simple_name in used.name():
                count += 1
                if count >= self.limiters[movement.simple_name]:
                    # print(f"Limiting {movement} because {used_so_far}")
                    return False
        return True


def simulate(steps, deck):
    ds = []
    for i, step in enumerate(steps):
        before_deck = deck
        deck = step.apply(deck)
        ds.append([i, step, before_deck, deck.tolist(), English().name_of_deck(deck)])
        print(i, deck)
    return pd.DataFrame(ds, columns=["i", "step", "before_deck", "after_deck", "new_mnemonica_simple_deck"])
    # print(name_of_deck(deck))

# def try_all_limits(deck, limits, divination):
#     print(f"\n\n4 ANY limit {limits} {divination}")
#     limit_search(AreKSetCardsTogether(4, all_card_sets_new_deck).check, deck=deck, limits=limits)



import numpy as np

from generativemagic.movement import Movement

DECK_TOP = 1
DECK_BOTTOM = 52


class Effect(Movement):
    def apply(self, sequence: np.array, chosen=None):
        raise Exception(f"Effect method not yet implemented for {self}")



def np_index(items, x):
    """Where is x in ar? This is a linear search.

    Parameters:
       items: numpy array to search for
       x: item to be found

    Returns:
       the position or -1 if not found"""

    for i in range(len(items)):
        if items[i] == x:
            return i
    return -1


def flat_map(xs):
    """Surprisingly computational efficient implementation for a flat map"""
    mapped = []
    for x in xs:
        mapped.extend(x)
    return mapped


def cards_positions(deck, cards):
    """Returns the card positions for all cards requested"""
    positions = [np_index(deck, card) for card in cards]
    return sorted(positions)



import numpy as np

from generativemagic.effect import Effect

REVERT_COMPLETELY = "REVERT_COMPLETELY"
ORIGINAL = "ORIGINAL"
REVERT_ROWS = "REVERT_ROWS"


class ThoughtTransference(Effect):
    """A NxN matrix based one card prediction."""

    def __init__(self, n, returning_order=None):
        self.__returning_order = returning_order
        if not self.__returning_order:
            self.__returning_order = REVERT_COMPLETELY
        self.__n = n

    def apply(self, sequence, chosen=None):

        if self.__returning_order == REVERT_COMPLETELY:
            return np.concatenate([sequence[self.__n ** 2 - 1::-1], sequence[self.__n ** 2:]])

        if self.__returning_order == ORIGINAL:
            return sequence

        rows = list(reversed([sequence[i * self.__n:(i + 1) * self.__n] for i in range(self.__n)]))
        rows.append(sequence[self.__n ** 2:])
        return np.concatenate(rows)

    def __repr__(self):
        return f"ThoughtTransference({self.__n},{self.__returning_order})"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return True



import logging
from typing import List

from generativemagic.arrays import np_index, cards_positions
from generativemagic.checks.basic import Checker


class IsGenericHand:
    def __init__(self, hands: List[int], can_second_deal: bool, desired_cards):
        self.__hands = hands
        self.__can_second_deal = can_second_deal
        self.__desired_cards = desired_cards

    def check(self, deck):
        for hand in self.__hands:
            found = self.__check_hand(deck, hand)
            if found:
                return hand, found
        return None

    def __check_hand(self, deck, hand):

        positions = cards_positions(deck, self.__desired_cards)

        if -1 in positions:
            # some of the cards were not found
            return None

        length = len(deck)

        for starting in positions:
            len_cards_to_count = len(self.__desired_cards)
            viable_positions = sorted([(starting + i * hand) % length for i in range(len_cards_to_count)])
            if positions == viable_positions:
                return positions

        if not self.__can_second_deal:
            return None

        # this simple O(hand) implementation does not support second dealing with any card, only with the ace
        # it also only supports the required N cards in the first N distributions
        # we can generate all combinations and make this a quick search

        # deltas between each cards
        deltas = [positions[i + 1] - positions[i] - 1 for i in range(len(positions) - 1)]

        # for each but last card
        for position in range(len(deltas) - 1):
            deltas[position] -= hand - 1
            if deltas[position] > 0:
                # too many cards, will have to hold a non ace, fail (could be done, of course)
                return None
            # holding the ace and removing from the next one...
            deltas[position + 1] += deltas[position]
            deltas[position] = 0

        # for the last card
        if deltas[-1] < 0:
            # missing cards before the last ace
            return None
        if deltas[-1] > hand - 1:
            # too many cards before the last one
            return None

        logging.info(positions)
        return positions, deltas


class IsAcesHand(IsGenericHand):
    def __init__(self, hands: List[int], can_second_deal):
        super().__init__(hands, can_second_deal, [1, 14, 27, 40])


class ArePokerHand(Checker):

    def __init__(self, desired, at_start=False):
        self._desired = desired
        self._at_start = at_start

    def check(self, deck):
        positions = cards_positions(deck, [0, 13, 26, 39])
        if self._at_start:
            if positions == [0, 4, 8, 12]:
                return positions
            return None
        l = len(deck)
        for starting in positions:
            if positions == sorted([starting, (starting + 4) % l, (starting + 8) % l, (starting + 12) % l]):
                return positions
        return None

    def __repr__(self):
        return f"ArePokerHand({self._desired},{self._at_start})"



from collections import defaultdict
from typing import List

from generativemagic.spelling import Language


class IsSpelled:
    """returns a card spelled from top"""

    def __init__(self, language: Language,
                 deltas: List[int],
                 supports_next: bool = False):
        self._supports_next = supports_next
        self._language = language
        self._deltas = deltas
        self._cached_lengths = defaultdict(set)
        for c in range(1, 53):
            name = language.card_name(c).replace(" ", "")
            self._cached_lengths[len(name)].add(c)
        self.items = self._cached_lengths.keys()

    def _instructions(self, card_number, direction: str, delta: int, deal: int, mention_middle: int, len_to_find,
                      position: int):
        # TODO discover why the int transformation is being required on numpy
        card_number = int(card_number)
        card_name = self._language.card_name(card_number)
        return card_number, direction, delta, card_name, deal, mention_middle, len_to_find, position

    def check(self, deck: List):
        # this or next card
        if self._supports_next:
            this_or_next = [0, 1]
        else:
            this_or_next = [0]

        # OF, no OF, OF with spaces
        mention_no_middle = [0, -2, 2]
        possibilities = []

        for len_to_find in self.items:
            for delta in self._deltas:
                for mention_middle in mention_no_middle:
                    for deal in this_or_next:

                        # position = len_to_find - 1 + delta + deal + mention_middle
                        position = len_to_find + delta + deal + mention_middle
                        card_number = deck[position]
                        if card_number in self._cached_lengths[len_to_find]:
                            instruct = self._instructions(card_number, "from top", delta, deal, mention_middle,
                                                          len_to_find, position)
                            possibilities.append(instruct)

                        position = -len_to_find - delta - deal - mention_middle
                        card_number = deck[position]
                        if card_number in self._cached_lengths[len_to_find]:
                            instruct = self._instructions(card_number, "from face", delta, deal, mention_middle,
                                                          len_to_find, position)
                            possibilities.append(instruct)

        if len(possibilities) == 0:
            return None
        return possibilities

    def __repr__(self):
        return f"IsSpelled({self._language})"



import numpy as np

from generativemagic.effect import Effect, DECK_TOP


class SagaciousJokerN1(Effect):
    """> Sagacious Joker N1, Jordan
> N cards are selected, their suit or color are revealed
> Sagacious Joker N2 is very similar
- Result: N cards plus the top card are displaced to the top or bottom
- Requires N people with pockets"""

    def __init__(self, selections: list, put_back_at):
        self._put_back_at = put_back_at
        self._selections = selections

    def apply(self, sequence, chosen=None):
        without = np.array([x for x in sequence if x not in self._selections])
        if self._put_back_at == DECK_TOP:
            return np.concatenate([self._selections, without])
        return np.concatenate([without, self._selections])

    def __repr__(self):
        return f"SagaciousJokerN1()"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return True



import numpy as np

SUIT_CLUBS = 0
SUIT_HEARTS = 1
SUIT_SPADES = 2
SUIT_DIAMONDS = 3
ACE = 1
TWO = 2
THREE = 3
FOUR = 4
FIVE = 5
SIX = 6
SEVEN = 7
EIGHT = 8
NINE = 9
TEN = 10
JACK = 11
QUEEN = 12
KING = 13

TWOS = [2, 2 + 13, 2 + 26, 2 + 39]
THREES = [3, 3 + 13, 3 + 26, 3 + 39]
FOURS = [4, 4 + 13, 4 + 26, 4 + 39]
FIVES = [5, 5 + 13, 5 + 26, 5 + 39]
SIXES = [6, 6 + 13, 6 + 26, 6 + 39]
EIGHTS = [8, 8 + 13, 8 + 26, 8 + 39]
NINES = [9, 9 + 13, 9 + 26, 9 + 39]
QUEENS = [12, 12 + 13, 12 + 26, 12 + 39]
KINGS = [13, 13 + 13, 13 + 26, 13 + 39]


def simple_deck():
    """A deck Ace to King. This is NOT the united statess based new deck order!"""
    return np.arange(1, 53)


def random_deck():
    """A random deck"""
    return np.random.permutation(52) + 1


def position_to_value(k: int) -> int:
    assert 0 < k < 53
    return (k - 1) % 13 + 1


def position_to_suit(k: int) -> int:
    assert 0 < k < 53
    return int((k - 1) / 13)


def position_to_suit_color(k: int):
    assert 0 < k < 53
    return int((k - 1) / 13) % 2


def value_and_suit_to_card(value: int, suit: int):
    assert 0 < value < 14 and 0 <= suit <= 3
    return suit * 13 + value


def value_and_suit_to_card_position_on_simple_deck(value: int, suit: int):
    return value_and_suit_to_card(value, suit) - 1


def is_int_card(card: int):
    return isinstance(card, (int, np.integer))


def deck_as_str(sequence):
    """Returns a deck as a string"""
    return ",".join(map(str, sequence))



from generativemagic.effect import Effect


class TwinSouls(Effect):
    """> TwinSouls, Al Baker
> Two cards are selected, two cards are revealed.
- Maintain order"""

    def apply(self, sequence, chosen=None):
        return sequence

    def __repr__(self):
        return f"TwinSouls()"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return True



from generativemagic.decks import value_and_suit_to_card, SUIT_CLUBS, SUIT_HEARTS, is_int_card
from generativemagic.effect import Effect

from generativemagic.spelling import Language, CachedLanguage, English

THREE_CLUBS = value_and_suit_to_card(3, SUIT_CLUBS)
THREE_HEARTS = value_and_suit_to_card(3, SUIT_HEARTS)
TWO_CLUBS = value_and_suit_to_card(2, SUIT_CLUBS)
REQUIRED_CARDS = [THREE_CLUBS, THREE_HEARTS, TWO_CLUBS]


class IsImpromptuSpelled:
    """returns if a SINGLE selected card can be spelled."""

    def __init__(self, can_second_deal: bool, language: Language):
        self._can_second_deal = can_second_deal
        self._language = language

    def check(self, deck, chosen=None):
        if not chosen:
            return None

        if not is_int_card(chosen):
            return None

        name = self._language.card_name(chosen)
        length = len(name.replace(" ", ""))

        # from top
        if deck[length - 1] == chosen:
            return f"Impromptu spell {name}, last letter is card"
        if deck[length] == chosen:
            return f"Impromptu spell {name}, next card is card"
        if self._can_second_deal and deck[length + 1] == chosen:
            return f"Second dealing last, impromptu spell {name}, next card is card"

        # from bottom
        if deck[-length - 1] == chosen:
            return f"Impromptu spell from bottom, {name}, next card is card"
        if deck[-length] == chosen:
            return f"Impromptu spell from bottom, {name}, last letter is card"

        return None

    def __repr__(self):
        return f"IsImpromptuSpelled({self._can_second_deal, self._language})"


class ImpromptuSpelled(Effect):

    def apply(self, sequence, chosen=None):
        pass

    def __repr__(self):
        return f"ImpromptuSpelled()"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return IsImpromptuSpelled(True, CachedLanguage(English())).check(deck, chosen)



from z3 import Int

from generativemagic.spelling import Language


def create_card_variables(language: Language):
    names = language.all_names()
    our_vars = map(Int, names)
    return list(our_vars), names



from typing import List, Counter

import pandas as pd
from tqdm import tqdm

from generativemagic.decks import position_to_value, deck_as_str


def all_compositions_of_fixed_length(decks: List, length: int,
                                     minimum_occurrences_per_deck: int = 1):
    compositions = Counter[str]()
    for deck in tqdm(decks):
        in_deck_occurrences = Counter[str]()
        for i in range(len(deck) - length + 1):
            card_numbers = [position_to_value(c) for c in deck[i:i + length]]
            sequence = deck_as_str(card_numbers)
            in_deck_occurrences[sequence] += 1
            if in_deck_occurrences[sequence] == minimum_occurrences_per_deck:
                compositions[sequence] += 1
    return compositions


def explore_card_combination(decks: List, lengths: List[int],
                             minimum_occurrences_per_deck: int = 1):
    results = {}
    for length in lengths:
        compositions = all_compositions_of_fixed_length(decks, length,
                                                        minimum_occurrences_per_deck=minimum_occurrences_per_deck)
        if len(compositions) == 0:
            continue
        df = pd.DataFrame.from_dict(compositions, orient='index').reset_index()
        df.columns = ["sequence", "frequency"]
        df['frequency'] /= len(decks)
        df.sort_values("frequency", ascending=False, inplace=True)
        results[length] = df
    return results



from numpy import argmin

from generativemagic.checks.poker import cards_positions
from generativemagic.decks import value_and_suit_to_card, SUIT_CLUBS, SUIT_HEARTS
from generativemagic.effect import Effect

THREE_CLUBS = value_and_suit_to_card(3, SUIT_CLUBS)
THREE_HEARTS = value_and_suit_to_card(3, SUIT_HEARTS)
TWO_CLUBS = value_and_suit_to_card(2, SUIT_CLUBS)
REQUIRED_CARDS = [THREE_CLUBS, THREE_HEARTS, TWO_CLUBS]


class IsTrucoHand:
    def __init__(self, hands, can_second_deal):
        self.__can_second_deal = can_second_deal
        self.__hands = hands

    def check(self, deck):
        for hand in self.__hands:
            found = self.__check_hand(deck, hand)
            if found:
                return hand, found
        return None

    def __check_hand(self, deck, hand):
        positions = cards_positions(deck, REQUIRED_CARDS)
        maximum_delta = hand * 3 + 1

        if self.__can_second_deal:
            maximum_delta += 1

        length = len(REQUIRED_CARDS)
        distances = []
        for i in range(length):
            first = positions[i]
            last = positions[(i + length - 1) % length]
            distance = last - first
            if distance < 0:
                distance += len(deck)
            distance += 1
            distances.append(distance)

        minimum_arg = argmin(distances)
        print(positions, maximum_delta, distances, minimum_arg)
        if distances[minimum_arg] <= maximum_delta:
            return {
                "minimum distance": distances[minimum_arg],
                "position": positions[minimum_arg],
                "positions": positions
            }

        return None


class TrucoHand(Effect):

    def apply(self, sequence):
        result = IsTrucoHand([2, 4], True).get_first_ace(sequence)
        raise Exception(f"To be implemented at {result}")

    def __repr__(self):
        return f"AcesHand()"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return IsTrucoHand([2, 4], True).check(deck)



import os
import pickle

from generativemagic.decks import simple_deck
from os import path


def load_movement(name):
    sequence = pickle.load(open(f"cached/movement_{name}.data", "rb"))
    return Reorder(name, sequence)


class Movement:

    def apply(self, current_stack):
        pass

    def reload(self, always_save: bool = True):
        if always_save or not path.isfile(self._get_cache_name()):
            self.save_result()
        return load_movement(self.__repr__())

    def still_can_use(self, used_so_far):
        return True

    def name(self):
        return self.__repr__()

    def save_result(self):
        result = self.apply(simple_deck() - 1)
        os.makedirs("cached", exist_ok=True)
        name = self._get_cache_name()
        pickle.dump(result, open(name, "wb"))

    def _get_cache_name(self):
        # bad looking, should not use repr..
        return f"cached/movement_{self.__repr__()}.data"


REPETITION_LIMIT = 2


class Reorder(Movement):

    def __init__(self, name, new_sequence):
        self.new_sequence = new_sequence
        self.internal_name = name
        self.simple_name = name[:name.index("(")]

    def __repr__(self):
        return f"Reorder({self.internal_name})"

    def apply(self, current_stack):
        return current_stack[self.new_sequence]

    def still_can_use(self, used_so_far):
        count = 0
        for used in used_so_far:
            if self.simple_name in used.name():
                count += 1
                if count >= REPETITION_LIMIT:
                    return False
        return True



import numpy as np

from generativemagic.decks import is_int_card
from generativemagic.effect import Effect, DECK_TOP


class IsSimplifiedReverse:
    def check(self, deck, chosen):
        if not is_int_card(chosen):
            return None
        return deck[0] == chosen or deck[-1] == chosen


class SimplifiedReverse(Effect):
    """A (chosen) card on top of the deck is found reversed in the middle of the deck."""

    def __init__(self, put_back_on):
        self.__put_back_on = put_back_on

    def apply(self, sequence, chosen=None):
        if self.__put_back_on == DECK_TOP:
            return sequence
        return np.roll(sequence, -1)

    def __repr__(self):
        return f"SimplifiedReverse({self.__put_back_on})"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return IsSimplifiedReverse().check(deck, chosen)



from collections import defaultdict
from functools import partial
from itertools import product
from multiprocessing import Pool
from typing import List, Type

import numpy as np
import pandas as pd
from tqdm import tqdm

from generativemagic.decks import simple_deck
from generativemagic.effect import Effect


class EffectInvalidParameterError(BaseException):
    pass


def explore_single(parameters, type_to_map: Type, deck_creator):
    try:
        instance: Effect = type_to_map(*parameters)
        result = instance.apply(deck_creator())
        return instance, result
    except EffectInvalidParameterError:
        return None


def run_parameter_space(type_to_map: Type, parameter_space, callback=None, deck_creator=simple_deck, threads=1):
    items = list(product(*parameter_space))
    results = []

    p = partial(explore_single, type_to_map=type_to_map, deck_creator=deck_creator)

    with Pool(threads) as pool:
        progress = tqdm(pool.imap(p, items), total=len(items))
        for pair in progress:
            if not pair:
                continue
            instance, result = pair
            if callback:
                callback(instance, result)
            results.append(result)
    return results


def collect_parameter_space(type_to_map, parameter_space):
    results = set()

    def appender(_, result):
        results.add(np.array2string(result, max_line_width=10000))

    run_parameter_space(type_to_map, parameter_space, appender)
    return results


def map_parameter_space(type_to_map, parameter_space, append_results_to=None):
    if append_results_to:
        results = append_results_to
    else:
        results = defaultdict(list)

    def appender(instance, result):
        results[np.array2string(result, max_line_width=10000)].append(instance.__repr__())

    run_parameter_space(type_to_map, parameter_space, appender)
    return results


def catalog_parameter_space():
    results = defaultdict(list)
    from generativemagic.effects.four_aces import parameter_space
    from generativemagic.effects.four_aces import FourAces
    from generativemagic.effects.guilhermesilveira.matrix_square import MathSquareMatrix
    from generativemagic.effects.hugard_encyclopedia.encyclopedia import TheMagicBreath
    from generativemagic.movements.simple_movements import TopToBottom
    from generativemagic.movements.simple_movements import BottomToTop

    effects = [[MathSquareMatrix, [[3, 4, 5, 6], [False, True], [False, True]]],
               # [ThreeMeetings, [range(10, 13), range(7, 10), [False, True]]],
               [TheMagicBreath, [range(5, 16), [False, True], [False, True]]],
               [FourAces, parameter_space()]]
    movements = [
        # [InFaro, []],
        #          [OutFaro, []],
        #          [CutAtScallop, [list(range(1, 53))]],
        [TopToBottom, [list(range(1, 27))]],
        [BottomToTop, [list(range(1, 27))]]]
    for t, p in (effects + movements):
        map_parameter_space(t, p, append_results_to=results)
    r = [(cards, types, len(types)) for cards, types in results.items()]
    return pd.DataFrame(r, columns=["cards", "types", "amount"])



import logging

from z3 import Solver

from generativemagic.solver.rules import Ruler
import timeit


class RuleSolver:

    def solve(self, ruler: Ruler):

        start_time = timeit.default_timer()
        solver = self._create_solver(ruler)

        print(f"Satisfiable? {solver.check()}")

        elapsed = timeit.default_timer() - start_time
        print(f"Elapsed {elapsed} seconds")
        print(solver.statistics())

        model = solver.model()
        if model:
            self._print_result(model)
        else:
            print("No model was found")

    def _print_result(self, model):
        dic_model = {}
        for d in model.decls():
            dic_model[int(model[d].__str__())] = d.name()
        print("{")
        keys = sorted(list(dic_model.keys()))
        for k in keys:
            print(f"{k}: '{dic_model[k]}',")
        print("}")

    def _create_solver(self, ruler):
        solver = Solver()
        unique_rules = set(ruler.rules())
        logging.info(f"Using {len(unique_rules)} unique constraints")
        for rule in unique_rules:
            solver.add(rule)
        return solver



from collections import defaultdict
from typing import Tuple, List

from generativemagic.decks import position_to_value, position_to_suit


def _matcher_starts(items, full_name: str):
    for i, item in enumerate(items):
        if item != "" and full_name.startswith(item):
            return i
    raise Exception(f"Did not find {full_name} in {items}")


def _matcher_ends(items, full_name: str):
    for i, item in enumerate(items):
        if full_name.endswith(item):
            return i
    raise Exception(f"Did not find {full_name} in {items}")


def _name_to_position(values, suits, name: str):
    value = _matcher_starts(values, name)
    suit = _matcher_ends(suits, name)
    return value + suit * 13


class Language:
    """Default structure for a basic language implementation in generative magic."""

    def card_name(self, c: int) -> str:
        raise Exception(f"not implemented on {type(self)}")

    def position_to_suit_name(self, k: int):
        raise Exception(f"not implemented on {type(self)}")

    def name_to_position(self, name: str):
        raise Exception(f"not implemented on {type(self)}")

    def name_of_deck(self, deck: List[int]):
        names = map(self.card_name, deck)
        return ",".join(names)

    def get_names_per_length(self):
        names = defaultdict(list)
        for c in range(1, 53):
            full_name = self.card_name(c)
            name = full_name.replace(" ", "")
            names[len(name)].append(full_name)
        return names

    def all_names(self):
        return list(map(self.card_name, range(1, 53)))


class CachedLanguage(Language):
    """Implementation that caches all card names."""

    def __init__(self, language: Language):
        self._language = language
        self._cards = [""]
        for c in range(1, 53):
            self._cards.append(language.card_name(c))

    def card_name(self, k: int):
        return self._cards[k]

    def name_to_position(self, name: str):
        return self._language.name_to_position(name)

    def position_to_suit_name(self, k: int):
        return self._language.position_to_suit_name(k)


class SimpleLanguage(Language):
    """A base implementation for most languages. It should be enough for several languages."""

    def __init__(self, names: Tuple, suits: Tuple):
        self._names = names
        self._suits = suits

    def position_to_suit_name(self, k):
        return self._suits[position_to_suit(k)]

    def name_to_position(self, name):
        return _name_to_position(self._names, self._suits, name)

    def card_name(self, c):
        return self.connect_value_and_suit(self._names[position_to_value(c)],
                                           self.position_to_suit_name(c))

    def connect_value_and_suit(self, value, suit):
        raise Exception(f"Not yet implemented for {self}")


class English(SimpleLanguage):
    def __init__(self):
        super().__init__(("", "ace", "two", "three", "four", "five",
                          "six", "seven", "eight", "nine", "ten",
                          "jack", "queen", "king"),
                         ("clubs", "hearts", "spades", "diamonds"))

    def connect_value_and_suit(self, value, suit):
        return f"{value} of {suit}"


class PortugueseCommon(SimpleLanguage):

    def __init__(self):
        super().__init__(("", "as", "dois", "tres", "quatro", "cinco", "seis", "sete", "oito", "nove", "dez",
                          "valete", "dama", "rei"),
                         ("paus", "coração", "espadas", "diamantes"))

    def connect_value_and_suit(self, value, suit):
        return f"{value} de {suit}"


class Portuguese(SimpleLanguage):

    def __init__(self):
        super().__init__(("", "as", "dois", "tres", "quatro", "cinco",
                          "seis", "sete", "oito", "nove", "dez",
                          "valete", "dama", "rei"),
                         ("paus", "copas", "espadas", "ouros"))

    def connect_value_and_suit(self, value, suit):
        return f"{value} de {suit}"


class French(SimpleLanguage):

    def __init__(self):
        super().__init__(("", "as", "deux", "trois", "quatre", "cinq",
                          "six", "sept", "huit", "neuf", "dix",
                          "valet", "dame", "roi"),
                         ("trèfles", "cœurs", "piques", "carreaux"))

    def connect_value_and_suit(self, value, suit):
        return f"{value} de {suit}"


class Spanish(SimpleLanguage):

    def __init__(self):
        super().__init__(("", "as", "dos", "tres", "cuatro", "cinco",
                          "seis", "siete", "ocho", "nueve", "diez",
                          "jota", "reina", "rei"),
                         ("tréboles", "corazones", "picas", "diamantes"))

    def connect_value_and_suit(self, value, suit):
        return f"{value} de {suit}"


class Japanese(SimpleLanguage):
    def __init__(self):
        super().__init__(("", "エース", "に", "さん", "よん", "ご",
                          "ろく", "なな", "はち", "きゅう", "じゅう",
                          "ジャック", "クイーン", "キング"),
                         ("クラブ", "ハート", "スペード", "ダイヤ"))

    def connect_value_and_suit(self, value, suit):
        return f"{suit}の{value}"


class Translator:
    def __init__(self, original: Language, target: Language):
        self._original = original
        self._target = target

    def translate(self, name: str):
        position = self._original.name_to_position(name)
        return self._target.card_name(position)

    def revert(self):
        return Translator(self._target, self._original)



import numpy as np

from generativemagic.effect import Effect, DECK_TOP


class InjectAt(Effect):
    """Injects a chosen cards at the top or bottom of the deck.
    This is a simple palming or putting the deck on top of the chosen cards.

    Why does it not allow to put anywhere? Because this is a simple movement to put on top or bottom."""

    def __init__(self, at: int):
        self.__at = at

    def apply(self, sequence, chosen=None):
        if self.__at == DECK_TOP:
            return np.insert(sequence, 0, chosen)
        return np.insert(sequence, len(sequence), chosen)

    def __repr__(self):
        return f"InjectAt({self.__at})"



import numpy as np

from random import shuffle


def concatenate(stacks):
    """Concatenates serveral stacks one on top of the other"""
    return np.concatenate(stacks)


def interweave(a, b):
    c = np.empty((a.size + b.size,), dtype=a.dtype)
    c[0::2] = a
    c[1::2] = b
    return c


def revert(deck):
    return np.flip(deck)


def stack_into(stack_count: int, card_count: int, deck, same_stack: bool = False):
    if same_stack:
        stacks = [deck[range(i * stack_count, i * stack_count + card_count)] for i in range(stack_count)]
    else:
        stacks = [deck[range(i, i + stack_count * card_count, card_count)] for i in range(stack_count)]
    remaining = deck[stack_count * card_count:]
    return stacks, remaining


def card_positions_on(deck, x):
    return [i + 1 for i, y in enumerate(deck) if y[:len(x)] == x]


def all_card_sets(deck):
    types = list(map(str, range(2, 10))) + ["ace", "jack", "queen", "king", "ten"]
    return [card_positions_on(deck, t) for t in types]


def all_card_sets_straight():
    return [[i, i + 13, i + 26, i + 39] for i in range(1, 14)]


def all_card_sets_new_deck():
    return [[i, i + 13, 53 - i - 13, 53 - i] for i in range(1, 14)]


def shuffled_copy(ar):
    m = list(ar)
    shuffle(m)
    return m



class Guilherme:
    """A proof of concept of an impromptu suggester for beginners to impromptu magic."""

    def __init__(self, known_effects):
        self.__known_effects = known_effects

    def suggestions_for(self, deck, chosen=None):
        return [effect for effect in self.__known_effects if effect.is_ready_to_use(deck, chosen)]



import logging
from typing import List

import z3
from z3 import Or, AtMost, Int

from generativemagic.solver.rules import Ruler, rules_aces_on_top, rules_all_cards_on_deck


class SpellingPositionRuler(Ruler):
    def __init__(self, all_vars, vars_per_length,
                 deltas_from_top=None,
                 deltas_from_bottom=None,
                 supports_next=False,
                 card_limiter=lambda c: True):
        self._supports_next = supports_next
        self._card_limiter = card_limiter
        self.__all_vars = all_vars
        self._positions = {}
        for i, card in enumerate(all_vars, start=1):
            self._positions[card] = i
        self.vars_per_length = vars_per_length
        if deltas_from_bottom is None:
            deltas_from_bottom = []
        self.deltas_from_bottom = deltas_from_bottom
        if deltas_from_top is None:
            deltas_from_top = []
        self.deltas_from_top = deltas_from_top
        self.__rules = set()
        self.used_positions = {1, 2, 3, 4}

    def create_rule(self, _, deck_order):
        all_ors = []
        for current_length in self.vars_per_length.keys():
            for original_card in self.vars_per_length[current_length]:
                if self._card_limiter and not self._card_limiter(self._positions[original_card]):
                    continue
                for delta in self.deltas_from_top:
                    self._position(deck_order[current_length - 1 + delta], all_ors, original_card)
                    if self._supports_next:
                        self._position(deck_order[current_length + delta], all_ors, original_card)
                for delta in self.deltas_from_bottom:
                    self._position(deck_order[- current_length - delta], all_ors, original_card)
                    if self._supports_next:
                        self._position(deck_order[- current_length - delta - 1], all_ors, original_card)
        all_ors = Or(all_ors)
        self.__rules.add(all_ors)

    def __add_rules_for_cards_must_be_in_deck(self, all_vars):
        self.__rules.update(rules_aces_on_top(all_vars))
        self.__rules.add(rules_all_cards_on_deck(all_vars))

        for c in self.used_positions:
            all_conditions = [card == c for card in all_vars]
            self.__rules.add(AtMost(*all_conditions, 1))
        logging.info(f"{len(self.used_positions)} used positions: {self.used_positions}")
        logging.info(f"{len(self.__rules)} generated rules")

    def add_final_rules(self, all_vars):
        self.__add_rules_for_cards_must_be_in_deck(all_vars)

    def rules(self):
        return list(self.__rules)

    def _position(self, card_number, all_ors: List, card: z3.Int):
        card_int = card_number.item()
        self.used_positions.add(card_int)
        all_ors.append(card == card_int)



from generativemagic.deck_operations import stack_into, concatenate, revert
from generativemagic.effect import Effect


class MathSquareMatrix(Effect):
    """The famous two matrix math magic. But with a few twists that allows better control of the deck."""

    def __init__(self, n: int, turn_after_first_split: bool = True, first_in_same_stack: bool = False,
                 deal_from_top=True, back_to_top=True):
        self.__back_to_top = back_to_top
        self.__n = n
        self.__turn_after_first_split = turn_after_first_split
        self.__first_in_same_stack = first_in_same_stack
        self.__deal_from_top = deal_from_top

    def apply(self, sequence):
        if not self.__deal_from_top:
            sequence = revert(sequence)

        stacks, remaining = stack_into(self.__n, self.__n, sequence, same_stack=self.__first_in_same_stack)
        stacks = concatenate(stacks)
        # fake shuffle
        if not self.__turn_after_first_split:
            stacks = revert(stacks)
        stacks, _ = stack_into(self.__n, self.__n, stacks)
        stacks = concatenate(stacks)

        if self.__back_to_top:
            remaining = revert(remaining)

        joined = concatenate([stacks, remaining])

        if not self.__deal_from_top and self.__back_to_top:
            joined = revert(joined)

        return joined

    def __repr__(self):
        return f"MathSquareMatrix({self.__n},turn_after_first_split={self.__turn_after_first_split},first_in_same_stack={self.__first_in_same_stack},deal_from_top={self.__deal_from_top},back_to_top={self.__back_to_top})"



import numpy as np

from generativemagic.arrays import np_index
from generativemagic.effect import DECK_TOP, DECK_BOTTOM
from generativemagic.movement import Movement


class ScallopCut(Movement):
    """Cuts to a scalloped, shorted, svengalied card. Simple, perfect cut to a card."""

    def __init__(self, scallop_card: int, cut_to: int = DECK_TOP):
        self._cut_to = cut_to
        self._scallop_card = scallop_card

    def apply(self, sequence, chosen=None):
        position = np_index(sequence, self._scallop_card)
        if self._cut_to == DECK_BOTTOM:
            return np.roll(sequence, len(sequence) - position - 1)
        return np.roll(sequence, -position)

    def __repr__(self):
        return f"ScallopCut({self._scallop_card})"



from generativemagic.decks import KING, QUEEN
from generativemagic.history.effects_history import Effects


def grade_five_or_ten(effects: Effects, message: str, accept_king=False, accept_queen=False):
    if effects.current_value() == 5:
        effects.append(1, "Rate 5 stars " + message)
        effects.append_if(5, "Out of 5 stars ")
    elif effects.current_value() == 10:
        effects.append(1, "Rate 10 " + message)
        effects.append_if(10, "Out of 10 ")
    elif accept_king and effects.current_value() == KING:
        effects.append(1, "They are the best, the king " + message)
        effects.append_if(KING, "Of all the kings")
        effects.append_if(QUEEN, "Or queens")
    elif accept_queen and effects.current_value() == QUEEN:
        effects.append(1, "They are the best, the queen " + message)
        effects.append_if(QUEEN, "Of all the queens")
        effects.append_if(KING, "Or kings")
    else:
        return False
    return True


def almost_perfect(effects: Effects, message: str):
    if effects.current_value() == 4:
        effects.append(1, "Rate 4 stars " + message)
        effects.append_if(5, "Out of 5 stars ")
    elif effects.current_value() == 9:
        effects.append(1, "Rate 9 " + message)
        effects.append_if(10, "Out of 10 ")
    else:
        return False
    return True


def perfect_or_imperfect(effects: Effects, msg: str):
    """Checks for either a perfect or near perfect result."""
    if grade_five_or_ten(effects, msg):
        return True
    return almost_perfect(effects, msg)


def generic_101(effects: Effects, type_learning: str):
    if effects.the_next_n_values([0, 1, 2, 3]) == [1, 1, 1, 1]:
        effects.append(4, f"{type_learning} learning one on one, not in the basic sense of one on one. But mentorship")
    if effects.the_next_n_values([0, 1]) == [1, 1]:
        effects.append(2, f"{type_learning} learning one on one")


def try_all_in_any_order(conditions):
    done = [False] * len(conditions)
    changed = True
    while changed:
        changed = False
        for i, condition in enumerate(conditions):
            if not done[i]:
                if condition():
                    done[i] = True
                    changed = True



import logging

import numpy as np
from z3 import Or, AtMost

from generativemagic.deck_operations import concatenate
from generativemagic.decks import SUIT_CLUBS, SUIT_HEARTS, SUIT_SPADES, SUIT_DIAMONDS, value_and_suit_to_card, \
    position_to_value
from generativemagic.effect import Effect
from generativemagic.solver.rules import Ruler, rules_aces_on_top


class TheMagicBreath(Effect):
    def __init__(self, n: int, put_stack_back_on_top=True, put_chosen_card_on_bottom=False):
        self.n = n
        self.put_stack_back_on_top = put_stack_back_on_top
        self.put_chosen_card_on_bottom = put_chosen_card_on_bottom

    def movements_description(self):
        return [["count", "n"], ["count", "n"],
                ["put back to bottom or top", "put_stack_back_on_top"],
                ["put_chosen_card_on_bottom", "put_chosen_card_on_bottom"]]

    def apply(self, sequence):
        if self.put_stack_back_on_top:

            if self.put_chosen_card_on_bottom:
                return np.roll(sequence, -1)

            return sequence

        if self.put_chosen_card_on_bottom:
            return concatenate([sequence[:self.n], sequence[self.n:], [sequence[0]]])

        return np.roll(sequence, -self.n)

    def __repr__(self):
        return f"TheMagicBreath({self.n},put_stack_back_on_top={self.put_stack_back_on_top},put_chosen_card_on_bottom={self.put_chosen_card_on_bottom})"


class PhenomenalThoughtCardsRuler(Ruler):
    """The Phenomenal Thought Cards effect allows the volunteer to pick a random card,
    put on the middle of the deck. The magician goes through the cards face down, until the magician says
    I have found it.

    That is not the chosen card, for example it is a 5. So now the magician says, well, five right?
    One, two, three, four, five (counts to the card), and there is the selected card.

    For this Ruler variation of the effect to work you need all number cards that can be used
    as counters to be marked because you don't know which one it will be until showtime. The possible
    cards is the *possible_cards* argument, which defaults to [7, 8 and 9]."""

    def __init__(self, all_vars, possible_cards=None):
        self.__all_vars = all_vars
        self.__rules = []
        if possible_cards:
            self.__possible_cards = possible_cards
        else:
            self.__possible_cards = [7, 8, 9]

    def create_rule(self, _, deck_order):
        conditions = []
        for value in self.__possible_cards:
            card_1_at_position = deck_order[-value]
            card_2_at_position = deck_order[value]
            cards = [value_and_suit_to_card(value, SUIT_CLUBS),
                     value_and_suit_to_card(value, SUIT_HEARTS),
                     value_and_suit_to_card(value, SUIT_SPADES),
                     value_and_suit_to_card(value, SUIT_DIAMONDS)]
            conditions.extend(list(map(lambda card: self.__all_vars[card] == card_1_at_position.item(), cards)))
            conditions.extend(list(map(lambda card: self.__all_vars[card] == card_2_at_position.item(), cards)))
        self.__rules.append(Or(conditions))

    def add_final_rules(self, all_vars):
        logging.info(f"{len(self.__rules)} generated rules")

    def rules(self):
        return self.__rules

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        for count in [7, 8, 9]:
            for position in [-count, count]:
                if position_to_value(deck[position]) == count:
                    return [[count], [position, position_to_value(deck[position]), deck[position]]]
        return None



import logging
from itertools import permutations
from typing import List

from generativemagic.decks import position_to_value, position_to_suit
from generativemagic.mapper import run_parameter_space
from generativemagic.spelling import Language


class Effects:

    def __init__(self, language: Language, sequence: List[int], effects=None, starting=0):
        self._language = language
        self._sequence = sequence
        if not effects:
            effects = []
        self._effects = effects
        self._current = starting
        self._update_values()

    def original_sequence(self):
        return self._sequence

    def _update_values(self):
        self._current_card = int(self._sequence[self._current])
        self._current_value = position_to_value(self._current_card)
        self._current_suit = position_to_suit(self._current_card)

    def current_value(self):
        return self._current_value

    def current_suit(self):
        return self._current_suit

    def move_to_next_card(self):
        if self.has_finished():
            return
        self._current += 1
        if self.has_finished():
            return
        self._update_values()

    def next_two_are(self, a, b):
        if self.is_out_of_bounds(1):
            return False
        following = position_to_value(self._sequence[self._current + 1])
        if self._current_value == a and following == b:
            return True
        if self._current_value == b and following == a:
            return True
        return False

    def next_are(self, *x):
        length = len(x)
        to_analyze = self._sequence[self._current:self._current + length]
        to_remove = list(x)
        for i in to_analyze:
            if i not in to_remove:
                return False
            to_remove.remove(i)
        return True

    def the_next_n_value(self, k):
        if self.is_out_of_bounds(k):
            return 10000
        return position_to_value(self._sequence[self._current + k])

    def the_next_n_suit(self, k):
        if self.is_out_of_bounds(k):
            return 10000
        return position_to_suit(self._sequence[self._current + k])

    def the_next_n_values(self, ks):
        if self.is_out_of_bounds(max(ks)):
            return [-1]
        return [self.the_next_n_value(k) for k in ks]

    def remaining_deck(self):
        return self._sequence[self._current:]

    def next_are_in_and_different(self, k, possible):
        so_far = set()
        for i in range(k):
            checking = self._current + i
            value = position_to_value(self._sequence[checking])
            if value in so_far:
                return False
            if value not in possible:
                return False
            so_far.add(value)
        return True

    def current_position(self):
        return self._current

    def copy(self):
        return Effects(self._language, self._sequence, self._effects.copy(), self._current)

    def description(self):
        return "\n".join(map(str, self._effects))

    def descriptionMD(self):
        return "- " + self.description().replace("\n", "\n- ")

    def append_if(self, card: int, text: str):
        if self.current_value() == card:
            self.append(1, text)
            return True
        return False

    def append(self, quantity, text):
        if self.has_finished():
            return
        pretext = []
        for i in range(quantity):
            pretext.append(f"{self._current_value}," + self._language.card_name(self._current_card))
            self.move_to_next_card()
        text = (pretext, text)
        self._effects.append(text)

    def append_map(self, conditions: dict):
        if self._current_value in conditions:
            self.append(1, conditions[self._current_value])
            return self._current_value
        return None

    def is_out_of_bounds(self, delta):
        return self._current + delta >= len(self._sequence)

    def has_finished(self):
        return self.is_out_of_bounds(0)

    def append_if_suit(self, suit, message):
        if self._current_suit == suit:
            self.append(1, message)
            return True
        return False

    def value_in(self, minimum, maximum):
        return minimum <= self.current_value() <= maximum

    def matches_values(self, values):
        if self.is_out_of_bounds(len(values)):
            return False
        for delta, value in enumerate(values):
            found = self._sequence[self._current + delta]
            if position_to_value(found) != value:
                return False
        return True


def all_possibilities_for(parts, minimum=0, limit_per_part=10000):
    possibilities = []
    max_parts = min(limit_per_part + 1, len(parts) + 1)
    for length in range(minimum, max_parts):
        possibilities.extend(list(permutations(parts, length)))
    logging.debug(f"Calculating for {parts}/{minimum}/{limit_per_part} got {len(possibilities)} possibilities")
    return possibilities


def extract_all_results_from_spelled(results):
    all_results = []
    for result in results:
        current = set()
        for item in result:
            current.add(item[3])
        all_results.extend(current)
    return all_results


def explore_results_of(main_effect, space, deck_creator, effect):
    results = run_parameter_space(main_effect, space, deck_creator=deck_creator)
    logging.info(results[0])
    logging.info(len(results[0]))

    results = list(map(effect.check, results))
    logging.info(results[0])
    return results



from collections import Iterable
from typing import List

from z3 import And, Or, Int


def rules_aces_on_top(all_vars: List[Int]):
    """Returns 4 rules so that all aces are on the top 4 positions"""

    aces = [all_vars[0], all_vars[13], all_vars[26], all_vars[39]]
    return list(map(rule_card_on_top_4, aces))


def rule_card_on_top_4(var_card: Int):
    """Returns one rule so that the card is on one of the top 4 positions"""

    # 24k => test_aces 234 secs
    return And(var_card >= 1, var_card <= 4)
    # 24k => test_aces(1) 1000 secs with OR!!!!!!
    # return Or(var_card == 1, var_card == 2, var_card == 3, var_card == 4)


def rules_all_cards_on_deck(all_vars: List[Int]):
    """Returns 52 cards must be on deck"""
    rules = []
    for var_card in all_vars:
        rules.append(var_card >= 1)
        rules.append(var_card <= 52)
    return And(rules)


class AllCardsOnDeck:
    """Left over as a non recommendation to use. Adding extra rules will make things slower.
    If they can get random values, that will be ok."""

    def __init__(self):
        self.__rules = []

    def rules(self) -> Iterable:
        return self.__rules

    def add_final_rules(self, all_vars: List[int]):
        self.__rules = [rules_all_cards_on_deck(all_vars)]


class Ruler:

    def rules(self) -> Iterable:
        pass

    def create_rule(self, x, deck):
        pass

    def add_final_rules(self, all_vars: List[int]):
        pass


class AndRuler(Ruler):
    def __init__(self, rulers):
        self.__rulers = rulers

    def create_rule(self, x, deck):
        for ruler in self.__rulers:
            ruler.create_rule(x, deck)

    def add_final_rules(self, all_vars):
        for ruler in self.__rulers:
            ruler.add_final_rules(all_vars)

    def rules(self):
        all_rules = []
        for ruler in self.__rulers:
            all_rules.extend(ruler.rules())
        return all_rules


def add_sequential_condition(firsts, seconds, all_vars):
    """all of the first must be followed by one of the seconds"""
    rules = []
    for first in firsts:
        possibles = []
        first_var = all_vars[first - 1]
        for second in seconds:
            second_var = all_vars[second - 1]
            possibles.append(first_var == (second_var + 1))
        rules.append(Or(possibles))
    return rules


def not_on_border(elements: List[int], all_vars,
                  can_be_on_first: bool = False,
                  can_be_on_last: bool = False):
    """all of the elements must not be in the border"""
    rules = []
    for element in elements:
        var = all_vars[element - 1]
        if not can_be_on_first:
            rules.append(var >= 2)
        if not can_be_on_last:
            rules.append(var <= 51)
    return rules


def on_deck(elements: List[int], all_vars):
    """all of the elements must be within the deck"""
    rules = []
    for element in elements:
        var = all_vars[element - 1]
        rules.append(var >= 1)
        rules.append(var <= 52)
    return rules



import pandas as pd


class ResultsFrequency:
    def frequency_analysis(self, elements):
        print("Value counts:")
        print(pd.value_counts(elements))
        print()
        print("Value counts normalized:")
        print(pd.value_counts(elements, normalize=True) * 100)
        print()
        unique_count = len(pd.unique(elements))
        print(f"Unique elements: {unique_count}")
        return pd.value_counts(elements).index



import numpy as np

from generativemagic.effect import Effect


class SecretAgent(Effect):
    """> Secret Agent, Guilherme Silveira
> Volunteer picks a number for the agent and does not say it (007).
> Counts the cards into a new stack. Peek the agent. Puts the stack back.
> Does a cut.
- Ends with the reverted order of the top N cards"""

    def __init__(self, codename: int):
        self._codename = codename

    def apply(self, sequence, chosen=None):
        return np.concatenate([sequence[self._codename - 1::-1],
                               sequence[self._codename:]])

    def __repr__(self):
        return f"SecretAgent({self._codename})"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        return True



from typing import List

import numpy as np

from generativemagic.effect import Effect


def riffle_generator(up_to: int, so_far: List = None):
    """Generates all combinations of TOP/BOTTOM/TOP/BOTTOM/etc card mixes up to a certain number"""
    if up_to == 0:
        return

    if not so_far:
        yield from riffle_generator(up_to, [0])
        so_far = []

    remaining = up_to - sum(so_far)
    if remaining == 0:
        yield so_far.copy()
        return

    for i in range(1, remaining + 1):
        so_far.append(i)
        yield from riffle_generator(up_to, so_far)
        so_far.pop()


def basic_parameter_space():
    return [list(riffle_generator(8))]


def parameter_space():
    return [list(riffle_generator(17))]


class RiffleShuffle(Effect):

    def __init__(self, interest: List[int]):
        self._interest = interest

    def apply(self, sequence: np.array, chosen=None):
        result = []
        is_right = 0
        positions = [0, 26]
        for amount in self._interest:
            at = positions[is_right]
            result.extend(sequence[at:at + amount])
            positions[is_right] += amount
            is_right = 1 - is_right
        return np.concatenate([result, sequence[positions[0]:26], sequence[positions[1]:]])

    def __repr__(self):
        return f"RiffleShuffle({self._interest})"



from collections import Counter
from functools import partial
from multiprocessing import Pool
from typing import List

from tqdm import tqdm

from generativemagic.history.effects_history import Effects, all_possibilities_for
from generativemagic.spelling import Language
import logging


class PartsStructure:
    def explore(self, all_effects: dict, anytime_parts: List, prune=52, limit_per_part=10000):
        raise Exception(f"Parts exploration must be implemented on {type(self)}")


class SimpleParts(PartsStructure):
    def __init__(self, parts, minimum=0, maximum=None):
        self._maximum = maximum
        self._minimum = minimum
        self._parts = parts

    def explore(self, all_effects: dict, anytime_parts: List, prune=52, limit_per_part=10000):
        # ugly, improve it
        if self._maximum:
            limit_per_part = self._maximum

        all_possible_stories = all_possibilities_for(self._parts, self._minimum, limit_per_part)

        temp_all_effects = self.create_temporary_list_of_effects_so_far(all_effects)

        largest_position = 1
        for effects in all_effects.values():

            # prune optimization
            if largest_position - effects.current_position() > prune:
                continue

            snapshot = effects.copy()
            for stories in all_possible_stories:
                effects = snapshot.copy()
                for story in stories:
                    for part in anytime_parts:
                        part(effects)
                    story(effects)
                    if effects.has_finished():
                        return effects

                # prune optimization
                if largest_position > effects.current_position() and \
                        largest_position - effects.current_position() > prune:
                    continue

                current = effects.current_position()

                exists = current in temp_all_effects
                if not exists:
                    temp_all_effects[current] = effects
                    largest_position = max(largest_position, current)

        return temp_all_effects

    def create_temporary_list_of_effects_so_far(self, all_effects):
        return all_effects.copy()


class AtLeastOnePart(SimpleParts):
    """Only effects that include at least one of these parts will be accepted"""

    def __init__(self, parts, minimum=1, maximum=None):
        super().__init__(parts, minimum, maximum)

    def create_temporary_list_of_effects_so_far(self, all_effects):
        return {}


class Optional(PartsStructure):
    def __init__(self, part):
        if type(part) is list:
            raise Exception("Only a single part can be optional")
        self._part = part

    def explore(self, all_effects: dict, anytime_parts: List[PartsStructure], prune=52, limit_per_part=10000):
        return SimpleParts([self._part]).explore(all_effects, anytime_parts, prune, limit_per_part)


class OrParts(PartsStructure):
    def __init__(self, parts):
        self._parts = parts

    def explore(self, all_effects: dict, anytime_parts: List, prune=52, limit_per_part=10000):
        for sub_parts in self._parts:
            result = _explore(sub_parts, anytime_parts, all_effects, prune, limit_per_part)
            if type(result) == Effects:
                return result
            if result != all_effects:
                return result
        return all_effects


def _explore(parts, anytimes_parts: List, all_effects: dict, prune: int = 52, limit_per_part: int = 10000):
    if type(parts) == list:
        return SimpleParts(parts).explore(all_effects, anytimes_parts, prune, limit_per_part)
    return parts.explore(all_effects, anytimes_parts, prune, limit_per_part)


class Parts:
    def __init__(self):
        self._parts = []

    def add(self, custom_part):
        self._parts.append(custom_part)
        return self

    def optional(self, part):
        return self.add(Optional(part))

    def mandatory(self, part):
        return self.add(AtLeastOnePart([part]))

    def parts(self) -> list:
        return self._parts.copy()

    def quantity(self, minimum, maximum, parts):
        return self.add(AtLeastOnePart(parts, minimum=minimum, maximum=maximum))

    def at_least_one(self, parts):
        return self.add(AtLeastOnePart(parts, minimum=1))


def explore_all(language: Language, sequence: List[int], all_parts: List,
                anytime_parts: List[PartsStructure],
                prune: int = 1000, limit_per_part=10) -> Effects:
    base_effects = Effects(language, sequence)
    all_effects = {0: base_effects}

    logging.info(f"Exploring all parts {all_parts}")

    for parts in all_parts:

        logging.debug(f"Going for parts {parts}")
        result = _explore(parts, anytime_parts, all_effects,
                          prune=prune, limit_per_part=limit_per_part)
        if type(result) == Effects:
            return result

        all_effects = result

    return find_best_effect(all_effects)


def find_best_effect(all_effects: dict) -> Effects:
    if type(all_effects) != dict:
        raise Exception(f"All effects should be a dict, got {all_effects}")
    max_target = max(all_effects.keys())
    effect = all_effects[max_target]
    return effect


def explore_single(my_deck, parts, anytime_parts, language: Language, prune=None):
    return explore_all(language, my_deck, parts, anytime_parts, prune=prune)


def explore_entire_results_history(decks, parts, anytime_parts,
                                   language: Language, prune=None, threads=1):
    failed = []
    total_count = len(decks)
    total_lacking = 0
    failed_count = 0
    failed_cards = []
    win_count = 0

    p = partial(explore_single, parts=parts, anytime_parts=anytime_parts, language=language, prune=prune)

    with Pool(threads) as pool:
        progress = tqdm(pool.imap(p, decks), total=total_count)
        for result in progress:
            to_go = total_count - win_count - failed_count
            win_message = f"Win {win_count} | Fail {failed_count} | To go {to_go}"
            progress.set_description(win_message)
            if result.has_finished():
                win_count += 1
            else:
                lacking = len(result.remaining_deck())
                total_lacking += lacking
                failed_count += 1
                failed.append(f"{result.remaining_deck()} - {lacking} - {result.original_sequence()}")
                print(result.original_sequence())
                print(result.description())
                failed_cards.append(result.current_value())

    failed_cards = Counter(failed_cards)
    failed = "\n".join(failed)
    print(f"% lacking: {total_lacking / total_count}")
    print(f"Success: {win_count}")
    print(f"Failed: {failed_count}")
    print(failed_cards)
    print(failed)



from itertools import permutations

import numpy as np

from generativemagic.deck_operations import concatenate
from generativemagic.effect import Effect
from generativemagic.mapper import EffectInvalidParameterError


def global_parameter_space():
    """This parameter space is unnecessarily too large"""
    return [list(range(7, 22)),
            list(range(12, 27)),
            list(range(25, 40)),
            list(permutations(range(1, 5)))]


def parameter_space():
    """This is a good parameter space as you spread the cards in your hands or go over them with your fingers.
    Remember to control the starting point by calling the next volunteer to say stop. The stop point by hiding 
    the cards with your other hand."""
    return [list(range(7, 17)),
            list(range(17, 27)),
            list(range(27, 37)),
            list(permutations(range(1, 5), 4))]


def basic_parameter_space2():
    return [list(range(9, 15)),
            list(range(15, 21)),
            list(range(21, 27)),
            list(permutations(range(1, 5), 4))]


def basic_parameter_space():
    return [list(range(9, 15)),
            list(range(15, 19)),
            list(range(19, 26)),
            list(permutations(range(1, 5), 4))]


class FourAces(Effect):
    """Can be improved by limiting the number of choices of the user, if required."""

    def __init__(self, cut_point_0, cut_point_1, cut_point_2, sorting_order):
        if cut_point_0 >= cut_point_1 or cut_point_1 >= cut_point_2:
            raise EffectInvalidParameterError
        self.cut_points = [cut_point_0, cut_point_1, cut_point_2]
        self.sorting_order = sorting_order

    def apply(self, sequence):
        p0 = sequence[:4]
        p1 = sequence[4:self.cut_points[0]]
        p2 = sequence[self.cut_points[0] + 1:self.cut_points[1]]
        p3 = sequence[self.cut_points[1] + 1:self.cut_points[2]]
        p4 = sequence[self.cut_points[2]:]
        p4 = concatenate([[sequence[self.cut_points[0]], sequence[self.cut_points[1]]], p4])
        ps = np.array([p0, p1, p2, p3, p4], dtype=object)

        # removed the 4 top cards
        return concatenate(ps[[*self.sorting_order]])

    def __repr__(self):
        return f"FourAces({self.cut_points},{self.sorting_order})"

    @staticmethod
    def is_ready_to_use(deck, chosen=None):
        aces_on_top = [card for card in deck[:4] if card in [0, 13, 26, 39]]
        if len(aces_on_top) == 4:
            return True
        return None



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="generativemagic",
    version="0.8.0",
    author="Guilherme Silveira",
    author_email="guilherme.silveira@gmail.com",
    description="Generative magic uses computer science, mathematics, design and statistics to generate new magic.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/guilhermesilveira/generativemagic",
    project_urls={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(include=["generativemagic*"]),
    python_requires=">=3.6",
    install_requires=["numpy>=1.20.1", "pandas>=1.2.3", "z3-solver>=4.8.10", "tqdm>=4.59"],
    tests_require=["pytest>=6.2.2"],
    test_suite='tests',
)



import numpy as np

from generativemagic.spelling import English

MNEMONICA_AS_STRING = np.array(['four of clubs', 'two of hearts', 'seven of diamonds', 'three of clubs',
                                'four of hearts', 'six of diamonds', 'ace of spades', 'five of hearts',
                                'nine of spades', 'two of spades', 'queen of hearts', 'three of diamonds',
                                'queen of clubs', 'eight of hearts', 'six of spades', 'five of spades',
                                'nine of hearts', 'king of clubs', 'two of diamonds', 'jack of hearts',
                                'three of spades', 'eight of spades', 'six of hearts', 'ten of clubs',
                                'five of diamonds', 'king of diamonds', 'two of clubs', 'three of hearts',
                                'eight of diamonds', 'five of clubs', 'king of spades', 'jack of diamonds',
                                'eight of clubs', 'ten of spades', 'king of hearts', 'jack of clubs',
                                'seven of spades', 'ten of hearts', 'ace of diamonds', 'four of spades',
                                'seven of hearts', 'four of diamonds', 'ace of clubs', 'nine of clubs',
                                'jack of spades', 'queen of diamonds', 'seven of clubs', 'queen of spades',
                                'ten of diamonds', 'six of clubs', 'ace of hearts', 'nine of diamonds']
                               )

MNEMONICA_AS_INTEGER = np.array(list(map(English().name_to_position, MNEMONICA_AS_STRING)))


def new_mnemonica_as_string():
    return MNEMONICA_AS_STRING.copy()


def new_mnemonica():
    return MNEMONICA_AS_INTEGER.copy()



from generativemagic.deck_operations import interweave
from generativemagic.movement import Movement


class InFaro(Movement):
    """In faro shuffle"""

    def apply(self, sequence):
        half = len(sequence) // 2
        return interweave(sequence[half:], sequence[:half])

    def __repr__(self):
        return f"InFaro()"


class OutFaro(Movement):
    """Out faro shuffle"""

    def apply(self, sequence):
        half = len(sequence) // 2
        return interweave(sequence[:half], sequence[half:])

    def __repr__(self):
        return f"OutFaro()"



import numpy as np

from generativemagic.movement import Movement


class TopToBottom(Movement):
    def __init__(self, n):
        self.n = n

    def apply(self, sequence, chosen=None):
        return np.roll(sequence, -self.n)

    def __repr__(self):
        return f"TopToBottom({self.n})"


class BottomToTop(Movement):
    def __init__(self, n):
        self.n = n

    def apply(self, sequence, chosen=None):
        return np.roll(sequence, self.n)

    def __repr__(self):
        return f"BottomToTop({self.n})"


class TopToLastToBottom(Movement):
    """Draw two cards from the deck in a simple motion, put them at the bottom of the deck."""

    def apply(self, sequence, chosen=None):
        return np.append(sequence[1:-1], [sequence[0], sequence[-1]])

    def __repr__(self):
        return f"TopToLastToBottom()"


class BottomToSecond(Movement):
    """Draw two cards from the deck in a simple motion, put them at the top of the deck."""

    def apply(self, sequence, chosen=None):
        return np.insert(sequence[:-1], 1, sequence[-1])

    def __repr__(self):
        return f"BottomToSecond()"


class ReceiveToTop(Movement):
    """Receives a card to the top of the stack."""

    def apply(self, sequence, chosen=None):
        return np.insert(sequence, 0, chosen)

    def __repr__(self):
        return f"ReceiveToTop()"


class ReceiveToBottom(Movement):
    """Receives a card to the bottom of the stack."""

    def apply(self, sequence, chosen=None):
        return np.concatenate([sequence, [chosen]])

    def __repr__(self):
        return f"ReceiveToBottom()"



class Checker:

    def check(self, deck):
        raise Exception("Not yet implemented")


class AreAll(Checker):
    """Checks if both of the checkers are valid"""

    def __init__(self, *all_checkers):
        self.all = all_checkers

    def check(self, deck):
        results = []
        for checker in self.all:
            result = checker.check(deck)
            if result is None:
                return None
            results.append(result)
        return results

    def __repr__(self):
        return f"AreAll({self.all})"


class AreAny(Checker):
    """Checks if any of the checkers are valid"""

    def __init__(self, *all_checkers):
        self.all = all_checkers

    def check(self, deck):
        for checker in self.all:
            result = checker.check(deck)
            if result:
                return result
        return None

    def __repr__(self):
        return f"AreAny({self.all})"
